@@ -7,9 +7,9 @@ Author: Jason A. Cox
 For more information see https://github.com/jasonacox/tinytuya
 
 Description
-    Setup Wizard will prompt the user for Tuya IoT Developer credentials and will gather all 
+    Setup Wizard will prompt the user for Tuya IoT Developer credentials and will gather all
     registered Device IDs and their Local KEYs.  It will save the credentials and the device
-    data in the tinytuya.json and devices.json configuration files respectively. The Wizard 
+    data in the tinytuya.json and devices.json configuration files respectively. The Wizard
     will then optionally scan the local devices for status.
 
     HOW to set up your Tuya IoT Developer account: iot.tuya.com:
@@ -22,28 +22,60 @@ Credits
     The TuyAPI/CLI wizard inspired and informed this python version.
 """
 # Modules
-import requests
-import time
+from __future__ import print_function
 import hmac
 import hashlib
+import ipaddress
 import json
-import pprint
-import logging
+import socket
+import time
+import requests
+from colorama import init
 import tinytuya
 
+# Optional libraries required for forced scanning
+try:
+    from getmac import get_mac_address
+    SCANLIBS = True
+except:
+    # Disable force scanning
+    SCANLIBS = False
+
 # Backward compatability for python2
 try:
     input = raw_input
 except NameError:
     pass
 
-def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorithm=True, body=None, headers=None):
+# Colorama terminal color capability for all platforms
+init()
+
+# Configuration Files
+DEVICEFILE = tinytuya.DEVICEFILE
+SNAPSHOTFILE = tinytuya.SNAPSHOTFILE
+CONFIGFILE = tinytuya.CONFIGFILE
+RAWFILE = tinytuya.RAWFILE
+
+# Global Network Configs
+DEFAULT_NETWORK = tinytuya.DEFAULT_NETWORK
+TCPTIMEOUT = tinytuya.TCPTIMEOUT    # Seconds to wait for socket open for scanning
+TCPPORT = tinytuya.TCPPORT          # Tuya TCP Local Port
+
+# Helper Functions
+def getmyIP():
+    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+    s.connect(("8.8.8.8", 80))
+    r = s.getsockname()[0]
+    s.close()
+    return r
+
+def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorithm=True, body=None, headers=None, version="1.0"):
     """Tuya IoT Platform Data Access
 
     Parameters:
         * region     Tuya API Server Region: us, eu, cn, in, us-e, eu-w
         * apiKey     Tuya Platform Developer ID
-        * apiSecret  Tuya Platform Developer secret 
+        * apiSecret  Tuya Platform Developer secret
         * uri        Tuya Platform URI for this call
         * token      Tuya OAuth Token
 
@@ -64,6 +96,7 @@ def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorit
         * Get Token = https://openapi.tuyaus.com/v1.0/token?grant_type=1
         * Get UserID = https://openapi.tuyaus.com/v1.0/devices/{DeviceID}
         * Get Devices = https://openapi.tuyaus.com/v1.0/users/{UserID}/devices
+        * Get Device info = https://openapi.tuyaus.com/v1.0/devices/factory-infos
 
     REFERENCE: https://images.tuyacn.com/smart/docs/python_iot_code_sample.py
 
@@ -71,36 +104,37 @@ def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorit
     # Set hostname based on apiRegion
     apiRegion = apiRegion.lower()
     urlhost = "openapi.tuyacn.com"          # China Data Center
-    if(apiRegion == "us"):
+    if apiRegion == "us":
         urlhost = "openapi.tuyaus.com"      # Western America Data Center
-    if(apiRegion == "us-e"):
+    if apiRegion == "us-e":
         urlhost = "openapi-ueaz.tuyaus.com" # Eastern America Data Center
-    if(apiRegion == "eu"):
+    if apiRegion == "eu":
         urlhost = "openapi.tuyaeu.com"      # Central Europe Data Center
-    if(apiRegion == "eu-w"):
+    if apiRegion == "eu-w":
         urlhost = "openapi-weaz.tuyaeu.com" # Western Europe Data Center
-    if(apiRegion == "in"):
+    if apiRegion == "in":
         urlhost = "openapi.tuyain.com"      # India Datacenter
 
     # Build URL
-    url = "https://%s/v1.0/%s" % (urlhost, uri)
+    url = "https://%s/v%s/%s" % (urlhost, version, uri)
 
     # Build Header
     now = int(time.time()*1000)
     headers = dict(list(headers.items()) + [('Signature-Headers', ":".join(headers.keys()))]) if headers else {}
-    if(token==None):
+    if token is None:
         payload = apiKey + str(now)
         headers['secret'] = apiSecret
     else:
         payload = apiKey + token + str(now)
-    
+
     # If running the post 6-30-2021 signing algorithm update the payload to include it's data
-    if new_sign_algorithm: payload += ('GET\n' +                                                                # HTTPMethod
-                                       hashlib.sha256(bytes((body or "").encode('utf-8'))).hexdigest() + '\n' + # Content-SHA256
-                                       ''.join(['%s:%s\n'%(key, headers[key])                                   # Headers
-                                                for key in headers.get("Signature-Headers", "").split(":")
-                                                if key in headers]) + '\n' +
-                                       '/' + url.split('//', 1)[-1].split('/', 1)[-1])  
+    if new_sign_algorithm:
+        payload += ('GET\n' +                                                                # HTTPMethod
+                    hashlib.sha256(bytes((body or "").encode('utf-8'))).hexdigest() + '\n' + # Content-SHA256
+                    ''.join(['%s:%s\n'%(key, headers[key])                                   # Headers
+                             for key in headers.get("Signature-Headers", "").split(":")
+                             if key in headers]) + '\n' +
+                    '/' + url.split('//', 1)[-1].split('/', 1)[-1])
     # Sign Payload
     signature = hmac.new(
         apiSecret.encode('utf-8'),
@@ -113,8 +147,8 @@ def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorit
     headers['sign'] = signature
     headers['t'] = str(now)
     headers['sign_method'] = 'HMAC-SHA256'
-    
-    if(token != None):
+
+    if token is not None:
         headers['access_token'] = token
 
     # Get Token
@@ -127,15 +161,16 @@ def tuyaPlatform(apiRegion, apiKey, apiSecret, uri, token=None, new_sign_algorit
         except:
             print("Failed to get valid JSON response")
 
-    return(response_dict)
+    return response_dict
 
-def wizard(color=True, retries=None):
+def wizard(color=True, retries=None, forcescan=False):
     """
     TinyTuya Setup Wizard Tuya based WiFi smart devices
 
     Parameter:
         color = True or False, print output in color [Default: True]
         retries = Number of retries to find IP address of Tuya Devices
+        forcescan = True or False, force network scan for device IP addresses
 
     Description
         Setup Wizard will prompt user for Tuya IoT Developer credentials and will gather all of
@@ -152,12 +187,8 @@ def wizard(color=True, retries=None):
         The TuyAPI/CLI wizard inspired and informed this python version.
     """
 
-    # Get Configuration Data
-    CONFIGFILE = 'tinytuya.json'
-    DEVICEFILE = 'devices.json'
-    RAWFILE = 'tuya-raw.json'
-    SNAPSHOTFILE = 'snapshot.json'
     config = {}
+    ip_list = {}
     config['apiKey'] = ''
     config['apiSecret'] = ''
     config['apiRegion'] = ''
@@ -170,23 +201,22 @@ def wizard(color=True, retries=None):
     except:
         # First Time Setup
         pass
-    
-    if(color == False):
-        # Disable Terminal Color Formatting
-        bold = subbold = normal = dim = alert = alertdim = ""
-    else:
-        # Terminal Color Formatting
-        bold = "\033[0m\033[97m\033[1m"
-        subbold = "\033[0m\033[32m"
-        normal = "\033[97m\033[0m"
-        dim = "\033[0m\033[97m\033[2m"
-        alert = "\033[0m\033[91m\033[1m"
-        alertdim = "\033[0m\033[91m\033[2m"
+
+    (bold, subbold, normal, dim, alert, alertdim, cyan, red, yellow) = tinytuya.termcolor(color)
 
     print(bold + 'TinyTuya Setup Wizard' + dim + ' [%s]' % (tinytuya.version) + normal)
     print('')
 
-    if(config['apiKey'] != '' and config['apiSecret'] != '' and
+    if forcescan:
+        if not SCANLIBS:
+            print(alert +
+                '    ERROR: force network scanning requested but not available - disabled.\n'
+                '           (Requires: pip install getmac)\n' + dim)
+            forcescan = False
+        else:
+            print(subbold + "    Option: " + dim + "Network force scanning requested.\n")
+
+    if (config['apiKey'] != '' and config['apiSecret'] != '' and
             config['apiRegion'] != '' and config['apiDeviceID'] != ''):
         needconfigs = False
         print("    " + subbold + "Existing settings:" + dim +
@@ -196,10 +226,10 @@ def wizard(color=True, retries=None):
         print('')
         answer = input(subbold + '    Use existing credentials ' +
                        normal + '(Y/n): ')
-        if(answer[0:1].lower() == 'n'):
+        if answer[0:1].lower() == 'n':
             needconfigs = True
 
-    if(needconfigs):
+    if needconfigs:
         # Ask user for config settings
         print('')
         config['apiKey'] = input(subbold + "    Enter " + bold + "API Key" + subbold +
@@ -230,7 +260,6 @@ def wizard(color=True, retries=None):
     SECRET = config['apiSecret']
     DEVICEID = config['apiDeviceID']
     REGION = config['apiRegion']        # us, eu, cn, in
-    LANG = 'en'                         # en or zh
 
     # Get Oauth Token from tuyaPlatform
     uri = 'token?grant_type=1'
@@ -242,7 +271,7 @@ def wizard(color=True, retries=None):
 
     token = response_dict['result']['access_token']
 
-    # Get UID from sample Device ID 
+    # Get UID from sample Device ID
     uri = 'devices/%s' % DEVICEID
     response_dict = tuyaPlatform(REGION, KEY, SECRET, uri, token)
 
@@ -255,14 +284,63 @@ def wizard(color=True, retries=None):
     # Use UID to get list of all Devices for User
     uri = 'users/%s/devices' % uid
     json_data = tuyaPlatform(REGION, KEY, SECRET, uri, token)
- 
-    # Filter to only Name, ID and Key
+
+    # Use Device ID to get MAC addresses
+    uri = 'devices/factory-infos?device_ids=%s' % (",".join(i['id'] for i in json_data['result']))
+    json_mac_data = tuyaPlatform(REGION, KEY, SECRET, uri, token)
+
+    if forcescan:
+        # Force Scan - Get list of all local ip addresses
+        try:
+            # Fetch my IP address and assume /24 network
+            ip = getmyIP()
+            network = ipaddress.IPv4Interface(u''+ip+'/24').network
+        except:
+            network = DEFAULT_NETWORK
+            ip = None
+            print(alert +
+                'ERROR: Unable to get your IP address and network automatically.\n'
+                '       (using %s)' % network + normal)
+
+        try:
+            # Warn user of scan duration
+            print("\n" + bold + "Scanning local network.  This may take a while..." + dim)
+            print(bold + '\n    Running Scan...' + dim)
+            # Loop through each host
+            for addr in ipaddress.IPv4Network(network):
+                # Fetch my IP address and assume /24 network
+                print(dim + '\r      Host: ' + subbold + '%s ...' % addr + normal, end='')
+                a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                a_socket.settimeout(TCPTIMEOUT)
+                location = (str(addr), TCPPORT)
+                result_of_check = a_socket.connect_ex(location)
+                if result_of_check == 0:
+                    # TODO: Verify Tuya Device
+                    ip = "%s" % addr
+                    mac = get_mac_address(ip=ip)
+                    ip_list[ip] = mac
+                    print(" Found Device [%s]" % mac)
+                a_socket.close()
+
+            print(dim + '\r      Done                           ' +normal +
+                        '\n\nDiscovered %d Tuya Devices\n' % len(ip_list))
+        except:
+            print('\n' + alert + '    Error scanning network - Ignoring' + dim)
+            forcescan = False
+
+    # Filter to only Name, ID and Key, IP and mac-address
     tuyadevices = []
     for i in json_data['result']:
         item = {}
         item['name'] = i['name'].strip()
         item['id'] = i['id']
         item['key'] = i['local_key']
+        try:
+            item['mac'] = next((m['mac'] for m in json_mac_data['result'] if m['id'] == i['id']), "N/A")
+            if forcescan:
+                item['ip'] = ip_list[item['mac']]
+        except:
+            pass
         tuyadevices.append(item)
 
     # Display device list
@@ -282,14 +360,14 @@ def wizard(color=True, retries=None):
         with open(RAWFILE, "w") as outfile:
             outfile.write(json.dumps(json_data, indent=4))
     except:
-        print('\n\n' + bold + 'Unable to save raw file' + dim )  
+        print('\n\n' + bold + 'Unable to save raw file' + dim )
 
     # Find out if we should poll all devices
     answer = input(subbold + '\nPoll local devices? ' +
                    normal + '(Y/n): ')
-    if(answer[0:1].lower() != 'n'):
+    if answer[0:1].lower() != 'n':
         # Set retries based on number of devices if undefined
-        if(retries == None):
+        if retries is None:
             retries = len(tuyadevices)+10+tinytuya.MAXCOUNT
 
         # Scan network for devices and provide polling data
@@ -302,7 +380,7 @@ def wizard(color=True, retries=None):
         def getIP(d, gwid):
             for ip in d:
                 if 'gwId' in d[ip]:
-                    if (gwid == d[ip]['gwId']):
+                    if gwid == d[ip]['gwId']:
                         return (ip, d[ip]['version'])
             return (0, 0)
 
@@ -317,14 +395,13 @@ def wizard(color=True, retries=None):
             item['ver'] = ver
             item['id'] = i['id']
             item['key'] = i['key']
-            if (ip == 0):
+            if ip == 0:
                 print("    %s[%s] - %s%s - %sError: No IP found%s" %
                       (subbold, name, dim, ip, alert, normal))
             else:
                 try:
                     d = tinytuya.OutletDevice(i['id'], ip, i['key'])
-                    if ver == "3.3":
-                        d.set_version(3.3)
+                    d.set_version(float(ver))
                     data = d.status()
                     if 'dps' in data:
                         item['dps'] = data
@@ -332,11 +409,11 @@ def wizard(color=True, retries=None):
                         try:
                             if '1' in data['dps'] or '20' in data['dps']:
                                 if '1' in data['dps']:
-                                        if data['dps']['1'] == True:
-                                            state = bold + "On" + dim
+                                    if data['dps']['1'] is True:
+                                        state = bold + "On" + dim
                                 if '20' in data['dps']:
-                                        if data['dps']['20'] == True:
-                                            state = bold + "On" + dim
+                                    if data['dps']['20'] is True:
+                                        state = bold + "On" + dim
                                 print("    %s[%s] - %s%s - %s - DPS: %r" %
                                     (subbold, name, dim, ip, state, data['dps']))
                             else:
@@ -356,7 +433,7 @@ def wizard(color=True, retries=None):
 
         # Save polling data snapsot
         current = {'timestamp' : time.time(), 'devices' : polling}
-        output = json.dumps(current, indent=4) 
+        output = json.dumps(current, indent=4)
         print(bold + "\n>> " + normal + "Saving device snapshot data to " + SNAPSHOTFILE)
         with open(SNAPSHOTFILE, "w") as outfile:
             outfile.write(output)
@@ -370,4 +447,4 @@ if __name__ == '__main__':
     try:
         wizard()
     except KeyboardInterrupt:
-        pass
\ No newline at end of file
+        pass
